"""
Serializers for country classes
"""

import gettext

import pycountry
from drf_yasg.utils import swagger_serializer_method
from geocurrency.core.helpers import validate_language
from pycountry import countries
from rest_framework import serializers

from .models import Country


class CountrySerializer(serializers.Serializer):
    """
    Serializer for Country
    """
    name = serializers.CharField(read_only=True)
    numeric = serializers.IntegerField(read_only=True)
    alpha_2 = serializers.CharField()
    alpha_3 = serializers.CharField(read_only=True)
    translated_name = serializers.SerializerMethodField()

    @staticmethod
    def validate_alpha2(alpha_2):
        """
        Validate that alpha 2 code is valid
        :param alpha_2: alpha 2 code from ISO3166
        """
        if countries.get(alpha_2=alpha_2):
            return alpha_2
        else:
            raise serializers.ValidationError('Invalid country alpha_2')

    def create(self, validated_data) -> Country:
        """
        Create a Country object
        :param validated_data: cleaned data
        """
        return Country(validated_data.get('alpha_2'))

    def update(self, instance, validated_data) -> Country:
        """
        Update a Country object
        :param instance: Contry object
        :param validated_data: cleaned data
        """
        country = Country(validated_data.get('alpha_2'))
        self.instance = country
        return self.instance

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_translated_name(self, obj: Country) -> str:
        """
        Translate country name
        :param obj: Country
        :return: translated name
        """
        request = self.context.get('request', None)
        if request:
            try:
                language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
                translation = gettext.translation('iso3166', pycountry.LOCALES_DIR,
                                                  languages=[language])
                translation.install()
                return translation.gettext(obj.name)
            except FileNotFoundError:
                return obj.name
        else:
            return obj.name


class CountryDetailSerializer(serializers.Serializer):
    """
    Detailed Serializer for Country
    """
    name = serializers.CharField(read_only=True)
    numeric = serializers.IntegerField(read_only=True)
    alpha_2 = serializers.CharField(read_only=True)
    alpha_3 = serializers.CharField(read_only=True)
    region = serializers.SerializerMethodField()
    subregion = serializers.SerializerMethodField()
    tld = serializers.SerializerMethodField()
    capital = serializers.SerializerMethodField()
    unit_system = serializers.SerializerMethodField()
    translated_name = serializers.SerializerMethodField()

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_region(self, obj: Country) -> str:
        """
        Country region wrapper
        :param obj: Country
        """
        return obj.region

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_subregion(self, obj: Country) -> str:
        """
        Country subregien wrapper
        :param obj: Country
        """
        return obj.subregion

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_tld(self, obj: Country) -> str:
        """
        Country TLD wrapper
        :param obj: Country
        """
        return obj.tld

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_capital(self, obj: Country) -> str:
        """
        Country capital wrapper
        :param obj: Country
        """
        return obj.capital

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_unit_system(self, obj: Country) -> str:
        """
        Country Unit system wrapper
        :param obj: Country
        """
        return obj.unit_system

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_translated_name(self, obj: Country) -> str:
        """
        Country translation wrapper
        :param obj: Country
        """
        request = self.context.get('request', None)
        if request:
            try:
                language = validate_language(request.GET.get('language', request.LANGUAGE_CODE))
                translation = gettext.translation('iso3166', pycountry.LOCALES_DIR,
                                                  languages=[language])
                translation.install()
                return translation.gettext(obj.name)
            except FileNotFoundError:
                return obj.name
        else:
            return obj.name



"""
Rates module tests
"""
import datetime
import uuid
from datetime import date

from django.conf import settings
from django.contrib.auth.models import User
from django.core.cache import cache
from django.test import TestCase
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient

from .models import Rate, RateConverter, NoRateFound
from .serializers import RateAmountSerializer


class RateTest(TestCase):
    """
    Rate object tests
    """
    base_currency = 'EUR'
    currency = 'USD'

    def setUp(self) -> None:
        """
        Setup test environment
        """
        settings.RATE_SERVICE = 'forex'
        self.user, created = User.objects.get_or_create(
            username='test',
            email='test@ipd.com'
        )
        self.user.set_password('test')
        self.user.save()
        Token.objects.create(user=self.user)
        self.key = uuid.uuid4()
        self.amounts = [
            {
                'currency': 'USD',
                'amount': 100,
                'date_obj': '2020-07-22'
            },
            {
                'currency': 'AUD',
                'amount': 50,
                'date_obj': '2020-07-23'
            },
        ]
        self.trash_amounts = [
            {
                'currency': 'USD',
                'amount': 'toto',
                'date_obj': '01/01/2020'
            },
            {
                'currency': 'LOL',
                'date_obj': '2020-07-23'
            },
            {
                'date_obj': '2020-07-23'
            },
            {
                'currency': 'JPY',
            },
        ]

    def test_fetch_rates(self):
        """
        Test fetching of rates from service
        """
        rates = Rate.objects.fetch_rates(base_currency=self.base_currency)
        self.assertIsNotNone(rates)

    def test_fetch_rates_with_date(self):
        """
        Test fetching of rates at a given date
        """
        rates = Rate.objects.fetch_rates(base_currency=self.base_currency,
                                         date_obj=date(year=2020, month=6, day=1))
        self.assertIsNotNone(rates)

    def test_fetch_rate(self):
        """
        Test fectch of a specific conversion rate
        """
        rate = Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        self.assertIsNotNone(rate)

    def test_fetch_rate_with_date(self):
        """
        Test of specific conversion rate at a given date
        """
        rate = Rate.objects.fetch_rates(
            base_currency=self.base_currency, currency=self.currency,
            date_obj=date(year=2020, month=6, day=1)
        )
        self.assertIsNotNone(rate)

    def test_find_rate(self):
        """
        Find a rate between currencies
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        rate = Rate.objects.find_rate(base_currency=self.base_currency, currency=self.currency)
        self.assertIsNotNone(rate, msg="no rate found")

    def test_find_pivot_rate(self):
        """
        Find rate between currencies that are not directly linked
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        Rate.objects.fetch_rates(base_currency=self.currency, currency='AUD')
        rate = Rate.objects.find_rate(base_currency=self.base_currency, currency='AUD')
        self.assertIsNotNone(rate, msg="no rate found")

    def test_rate_at_date(self):
        """
        Find rate at a specific date
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        Rate.objects.fetch_rates(base_currency=self.currency, currency='AUD')
        rate = Rate.objects.find_rate(base_currency=self.base_currency, currency=self.currency)
        self.assertIsNotNone(rate.pk, msg="no direct rate found")
        rate = Rate.objects.find_rate(base_currency=self.base_currency, currency='AUD')
        self.assertIsNotNone(rate.pk, msg="no pivot rate found")

    def test_custom_rate(self):
        """
        Create a custom rate
        """
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AUD',
            base_currency='AFN',
            value=0.123,
            value_date='2021-01-01'
        )
        self.assertIsNotNone(
            Rate.objects.filter(user=self.user, key=self.key, base_currency='AFN', currency='AUD'))

    def test_find_rate_chain(self):
        """
        Test find rate with keys
        """
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AUD',
            base_currency='AFN',
            value=0.123,
            value_date='2021-01-01'
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='JPY',
            value=200,
            value_date='2021-01-01'
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='JPY',
            base_currency='BND',
            value=13,
            value_date='2021-01-01'
        )
        rates = Rate.objects.currency_shortest_path(
            currency='AUD', base_currency='BND', key=self.key, date_obj='2021-01-01')
        self.assertEqual(rates, ['AUD', 'AFN', 'JPY', 'BND'])

    def test_find_mixed_rate_chain(self):
        """
        Test find rate with keys with gaps in the key chain
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency)
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='ARS',
            base_currency='AFN',
            value=0.123,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='BND',
            value=200,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='BND',
            base_currency='JPY',
            value=13,
            value_date=datetime.date.today()
        )
        rates = Rate.objects.currency_shortest_path(
            currency='ARS', base_currency='EUR', key=self.key, date_obj=datetime.date.today())
        self.assertEqual(rates, ['ARS', 'AFN', 'BND', 'JPY', 'EUR'])

    def test_no_rate_path(self):
        """
        Test what happens if no rate is found
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency)
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='ARS',
            base_currency='AFN',
            value=0.123,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='BND',
            value=200,
            value_date=datetime.date.today()
        )
        self.assertRaises(
            NoRateFound,
            Rate.objects.currency_shortest_path,
            currency='ARS',
            base_currency='EUR',
            key=self.key,
            date_obj=datetime.date.today()
        )

    def test_find_rate_custom(self):
        """
        Test conversion between rates with custom rate
        """
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='ARS',
            base_currency='AFN',
            value=0.123,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='BND',
            value=200,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='BND',
            base_currency='JPY',
            value=13,
            value_date=datetime.date.today()
        )
        rate = Rate.objects.find_rate(
            base_currency='JPY',
            currency='ARS',
            date_obj=datetime.date.today(),
            key=self.key)
        self.assertEqual(rate.value, 0.123 * 200 * 13)

    def test_find_rate_mixed(self):
        """
        Test conversion between rates with custom and standard rates
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency)
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='ARS',
            base_currency='AFN',
            value=0.123,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='BND',
            value=200,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='BND',
            base_currency='JPY',
            value=13,
            value_date=datetime.date.today()
        )
        jpy_eur = Rate.objects.get(
            base_currency='EUR',
            currency='JPY',
            value_date=datetime.date.today()
        )
        rate = Rate.objects.find_rate(
            base_currency='EUR',
            currency='ARS',
            date_obj=datetime.date.today(),
            key=self.key)
        self.assertEqual(rate.value, 0.123 * 200 * 13 * jpy_eur.value)

    def test_find_rate_override(self):
        """
        Test having 2 rates for the same period, with a custom and a standard one
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency)
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='ARS',
            base_currency='AFN',
            value=0.123,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='AFN',
            base_currency='BND',
            value=200,
            value_date=datetime.date.today()
        )
        Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='BND',
            base_currency='JPY',
            value=13,
            value_date=datetime.date.today()
        )
        jpy_eur_custom = Rate.objects.create(
            user=self.user,
            key=self.key,
            currency='JPY',
            base_currency='EUR',
            value=100,
            value_date=datetime.date.today()
        )
        jpy_eur = Rate.objects.get(
            base_currency='EUR',
            currency='JPY',
            user__isnull=True,
            value_date=datetime.date.today()
        )
        rate = Rate.objects.find_rate(
            base_currency='EUR',
            currency='ARS',
            date_obj=datetime.date.today(),
            key=self.key)
        self.assertEqual(rate.value, 0.123 * 200 * 13 * jpy_eur_custom.value)


class RateAPITest(TestCase):
    """
    Rate object tests
    """
    base_currency = 'EUR'
    currency = 'USD'

    def setUp(self) -> None:
        """
        Setup test environment
        """
        settings.RATE_SERVICE = 'forex'
        self.user, created = User.objects.get_or_create(
            username='test',
            email='test@ipd.com'
        )
        self.user.set_password('test')
        self.user.save()
        Token.objects.create(user=self.user)
        self.key = uuid.uuid4()
        self.amounts = [
            {
                'currency': 'USD',
                'amount': 100,
                'date_obj': '2020-07-22'
            },
            {
                'currency': 'AUD',
                'amount': 50,
                'date_obj': '2020-07-23'
            },
        ]
        self.trash_amounts = [
            {
                'currency': 'USD',
                'amount': 'toto',
                'date_obj': '01/01/2020'
            },
            {
                'currency': 'LOL',
                'date_obj': '2020-07-23'
            },
            {
                'date_obj': '2020-07-23'
            },
            {
                'currency': 'JPY',
            },
        ]

    def test_post_rate(self):
        """
        Test rate creation
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(response.status_code, 201)
        response = client.get(
            '/rates/',
            format='json')
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 2)
        else:
            self.assertEqual(len(response.json()), 2)

    def test_post_rate_without_key(self):
        """
        Test standard rate creation without key
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.post(
            '/rates/',
            data={
                'key': '',
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_request(self):
        """
        Test rate list
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        client = APIClient()
        response = client.get(
            '/rates/',
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_stats_request(self):
        """
        Test rate stats
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        client = APIClient()
        response = client.get(
            '/rates/stats/',
            data={},
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_connected_list_request(self):
        """
        Test rates list for authenticated user
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        self.assertIn('id', post_response.json())
        response = client.get(
            '/rates/',
            format='json')
        anon_client = APIClient()
        anon_response = anon_client.get(
            '/rates/',
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(anon_response.status_code, status.HTTP_200_OK)
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']),
                             len(anon_response.json()['results']) + 2)
        else:
            self.assertEqual(len(response.json()), len(anon_response.json()) + 2)

    def test_list_with_key_request(self):
        """
        Test list with key argument
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        response = client.get(
            '/rates/',
            data={'key': self.key},
            format='json')
        if 'results' in response.json():
            self.assertEqual(response.json()['results'][0]['key'], str(self.key))
        else:
            self.assertEqual(response.json()[0]['key'], str(self.key))

    def test_list_with_key_or_null_request(self):
        """
        Test key_or_null filter
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        response = client.get(
            '/rates/',
            data={'key_or_null': self.key},
            format='json')
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 3)
        else:
            self.assertEqual(len(response.json()), 3)

    def test_list_with_key_isnull_request(self):
        """
        Test key_isnull filter
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        response = client.get(
            '/rates/',
            data={'key_isnull': self.key},
            format='json')
        if 'results' in response.json():
            self.assertEqual(response.json()['results'][0]['key'], None)
        else:
            self.assertEqual(response.json()[0]['key'], None)

    def test_list_with_key_and_currency_request(self):
        """
        Test key and currency filters
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        response = client.get(
            '/rates/',
            data={'key': self.key, 'currency': 'USD'},
            format='json')
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 1)
        else:
            self.assertEqual(len(response.json()), 1)

    def test_stats_with_key_and_currency_request(self):
        """
        Test stats with key and currency filters
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-02',
                'value': 1.20,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        response = client.get(
            '/rates/stats/',
            data={'key': self.key, 'currency': 'EUR'},
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_with_key_and_base_currency_request(self):
        """
        Test stats with key and base_currency filters
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'value_date': '2020-01-01',
                'value': 1.10,
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        response = client.get(
            '/rates/',
            data={'key': self.key, 'base_currency': 'USD'},
            format='json')
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 1)
        else:
            self.assertEqual(len(response.json()), 1)

    def test_retrieve_request(self):
        """
        Test retreiving a specific rate
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        rate = Rate.objects.first()
        client = APIClient()
        response = client.get(
            f'/rates/{rate.pk}/',
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], rate.pk)

    def test_bulk_create_request(self):
        """
        Test of bulk creation
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/bulk/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'from_date': '2020-01-01',
                'to_date': '2020-09-01',
                'value': 1.10
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(len(post_response.json()),
                         (datetime.date(year=2020, month=9, day=1) - datetime.date(year=2020,
                                                                                   month=1,
                                                                                   day=1)).days + 1)

    def test_latest_currency_request(self):
        """
        Test filtering latest values for currency
        """
        client = APIClient()
        token = Token.objects.get(user__username=self.user.username)
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        post_response = client.post(
            '/rates/bulk/',
            data={
                'key': self.key,
                'currency': 'USD',
                'base_currency': 'EUR',
                'from_date': '2020-01-01',
                'to_date': '2020-09-01',
                'value': 1.10
            }
        )
        self.assertEqual(post_response.status_code, status.HTTP_201_CREATED)
        response = client.get(
            '/rates/?currency_latest_values=USD'
        )
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 1)
        else:
            self.assertEqual(len(response.json()), 1)
        response = client.get(
            '/rates/?base_currency_latest_values=EUR'
        )
        if 'results' in response.json():
            self.assertEqual(len(response.json()['results']), 1)
        else:
            self.assertEqual(len(response.json()), 1)


class RateConverterTest(TestCase):
    """
    Test Rate conversion
    """

    base_currency = 'EUR'
    currency = 'USD'

    def setUp(self) -> None:
        """
        setup test environment
        """
        settings.RATE_SERVICE = 'forex'
        from_date = datetime.date(year=2020, month=7, day=20)
        to_date = datetime.date(year=2020, month=7, day=25)
        Rate.objects.fetch_rates(base_currency=self.base_currency, date_obj=from_date,
                                 to_obj=to_date)
        self.user, created = User.objects.get_or_create(
            username='test',
            email='test@ipd.com'
        )
        self.converter = RateConverter(user=self.user, base_currency='EUR')
        self.amounts = [
            {
                'currency': 'USD',
                'amount': 100,
                'date_obj': '2020-07-22'
            },
            {
                'currency': 'AUD',
                'amount': 50,
                'date_obj': '2020-07-23'
            },
        ]
        self.trash_amounts = [
            {
                'currency': 'USD',
                'amount': 'toto',
                'date_obj': '01/01/2020'
            },
            {
                'currency': 'LOL',
                'date_obj': '2020-07-23'
            },
            {
                'date_obj': '2020-07-23'
            },
            {
                'currency': 'JPY',
            },
        ]

    def test_created(self):
        """
        Test converter initial status
        """
        self.assertEqual(self.converter.status, self.converter.INITIATED_STATUS)

    def test_add_data(self):
        """
        Test adding data to a converter
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        errors = self.converter.add_data(self.amounts)
        self.assertEqual(errors, [])
        self.assertEqual(self.converter.status, self.converter.INSERTING_STATUS)
        self.assertIsNotNone(self.converter.cached_currencies)
        self.assertIsNotNone(cache.get(self.converter.id))

    def test_trash_amounts(self):
        """
        Test adding trash data to the converter
        """
        converter = RateConverter(user=self.user, base_currency='EUR')
        errors = converter.add_data(self.trash_amounts)
        self.assertEqual(len(errors), 4)
        self.assertIn("amount", errors[0])
        self.assertIn("currency", errors[1])
        self.assertNotIn("date_obj", errors[2])
        self.assertNotIn("currency", errors[3])

    def test_convert(self):
        """
        Conversion test
        """
        result = self.converter.convert()
        self.assertEqual(result.id, self.converter.id)
        self.assertEqual(result.target, 'EUR')
        self.assertEqual(self.converter.status, self.converter.FINISHED)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.detail), len(self.converter.data))
        converted_sum = sum([d.converted_value for d in result.detail])
        self.assertEqual(result.sum, converted_sum)

    def test_convert_pivot(self):
        """
        Test converting currencies with indirect relation
        """
        Rate.objects.fetch_rates(base_currency='EUR', currency='AUD',
                                 date_obj=datetime.date(year=2020, month=7, day=23))
        Rate.objects.fetch_rates(base_currency='JPY', currency='EUR',
                                 date_obj=datetime.date(year=2020, month=7, day=23))
        converter = RateConverter(self.user, base_currency='JPY')
        amounts = [
            {
                'currency': 'AUD',
                'amount': 50,
                'date_obj': '2020-07-23'
            },
        ]
        converter.add_data(amounts)
        result = converter.convert()
        self.assertEqual(result.id, converter.id)
        self.assertEqual(result.target, 'JPY')
        self.assertEqual(converter.status, converter.FINISHED)
        self.assertEqual(result.errors, [])
        self.assertEqual(len(result.detail), len(converter.data))
        converted_sum = sum([d.converted_value for d in result.detail])
        self.assertEqual(result.sum, converted_sum)


class RateConverterAPITest(TestCase):
    """
    Test Rate conversion API
    """

    base_currency = 'EUR'
    currency = 'USD'

    def setUp(self) -> None:
        """
        setup test environment
        """
        settings.RATE_SERVICE = 'forex'
        from_date = datetime.date(year=2020, month=7, day=20)
        to_date = datetime.date(year=2020, month=7, day=25)
        Rate.objects.fetch_rates(base_currency=self.base_currency, date_obj=from_date,
                                 to_obj=to_date)
        self.user, created = User.objects.get_or_create(
            username='test',
            email='test@ipd.com'
        )
        self.converter = RateConverter(user=self.user, base_currency='EUR')
        self.amounts = [
            {
                'currency': 'USD',
                'amount': 100,
                'date_obj': '2020-07-22'
            },
            {
                'currency': 'AUD',
                'amount': 50,
                'date_obj': '2020-07-23'
            },
        ]
        self.trash_amounts = [
            {
                'currency': 'USD',
                'amount': 'toto',
                'date_obj': '01/01/2020'
            },
            {
                'currency': 'LOL',
                'date_obj': '2020-07-23'
            },
            {
                'date_obj': '2020-07-23'
            },
            {
                'currency': 'JPY',
            },
        ]

    def test_convert_request(self):
        """
        Test conversion request
        """
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        Rate.objects.fetch_rates(base_currency=self.currency, currency='AUD')
        amounts = RateAmountSerializer(self.amounts, many=True)
        client = APIClient()
        response = client.post(
            '/rates/convert/',
            data={
                'data': amounts.data,
                'target': 'EUR',
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('sum', response.json())
        self.assertEqual(len(response.json().get('detail')), len(self.amounts))

    def test_convert_batch_request(self):
        """
        Test batch conversion
        """
        batch_id = uuid.uuid4()
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        Rate.objects.fetch_rates(base_currency=self.currency, currency='AUD')
        client = APIClient()
        amounts = RateAmountSerializer(self.amounts, many=True)
        response = client.post(
            '/rates/convert/',
            data={
                'data': amounts.data,
                'target': 'EUR',
                'batch_id': batch_id,
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'), RateConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.post(
            '/rates/convert/',
            data={
                'data': amounts.data,
                'batch_id': batch_id,
                'target': 'EUR',
                'eob': True
            },
            format='json')
        self.assertEqual(response.json().get('status'), RateConverter.FINISHED)
        self.assertEqual(len(response.json().get('detail')), 2 * len(self.amounts))

    def test_watch_request(self):
        """
        Test observation of the batch
        """
        batch_id = uuid.uuid4()
        Rate.objects.fetch_rates(base_currency=self.base_currency, currency=self.currency)
        Rate.objects.fetch_rates(base_currency=self.currency, currency='AUD')
        client = APIClient()
        amounts = RateAmountSerializer(self.amounts, many=True)
        response = client.post(
            '/rates/convert/',
            data={
                'data': amounts.data,
                'target': 'EUR',
                'batch_id': batch_id,
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('id', response.json())
        self.assertEqual(response.json().get('status'), RateConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))
        response = client.get(
            f'/watch/{str(batch_id)}/',
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json().get('status'), RateConverter.INSERTING_STATUS)
        self.assertEqual(response.json().get('id'), str(batch_id))



"""
REST permissions
"""

from rest_framework import permissions


class CountryObjectPermission(permissions.BasePermission):
    """
    RO Permissions for Country objects
    """

    def has_object_permission(self, request, view, obj) -> bool:
        """
        check permission
        """
        if request.method in permissions.SAFE_METHODS:
            return True
        return False



"""
Core helpers
"""
from importlib import import_module

import logging
from django.apps import apps
from django.conf import settings


def service(service_type: str, service_name: str, *args, **kwargs):
    """
    Service wrapper
    Services are used to abstract calls to external services
    """
    service_path = settings.SERVICES[service_type][service_name]
    if not hasattr(apps, 'services'):
        setattr(apps, 'services', {})
    if service_name in apps.services:
        return apps.services.get(service_name)
    try:
        service = service_path.split('.')
        module_name = '.'.join(service[:-1])
        class_name = service[-1]
        module = import_module(module_name)
        service_class = getattr(module, class_name)
        service = service_class(*args, **kwargs)
        apps.services[service_name] = service
        return service
    except (AttributeError, ImportError, KeyError) as e:
        logging.error(e)
        return None


def validate_language(lang):
    """
    Validate languages based on settings
    """
    if lang in [l[0] for l in settings.LANGUAGES]:
        return lang
    return 'en'


"""
Settings specific to Country module
"""

FLAG_SOURCE = 'https://raw.githubusercontent.com/cristiroma/countries/master/data/flags/SVG/{alpha_2}.svg?sanitize=true'

# put in global settings.py to override
GEOCODING_SERVICE = 'pelias'

# put in global settings.py to override
GEOCODER_PELIAS_URL = 'https://api.geocode.earth/v1'
GEOCODER_PELIAS_KEY = ''
GEOCODER_GOOGLE_KEY = ''

GEOCODING_SERVICE_SETTINGS = {
    'pelias': {
        'class': 'geocurrencies.countries.services.pelias',
        'default_url': 'http://127.0.0.1:3100/v1',
        'reverse': {
            'service': 'reverse',
        },
        'search': {
            'service': 'search',  # path of the search service
            'address': 'text'
        }

    },
    'google': {
        'class': 'geocurrencies.countries.services.google',
        'default_url': 'https://maps.googleapis.com/maps/api',
        'reverse': {
            'service': 'reverse/json',
        },
        'search': {
            'service': 'geocode/json',  # Path of the search service
            'address': 'address'
        }
    }

}



"""
Core views
"""

from django.conf import settings
from django.shortcuts import render
from django.views.decorators.http import require_safe
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.views import APIView


@require_safe
def index(request):
    """
    Base page
    """
    return render(request, 'index.html')


@require_safe
def tos(request):
    """
    Term of Service view
    """
    return render(request, 'tos.html')


class LanguageView(APIView):
    """
    Language view
    """

    @action(['GET'], detail=False, url_path='languages', url_name="languages")
    def get(self, request, *args, **kwargs):
        """
        GET handler
        :param request: HTTPRequest
        """
        return Response([{'name': l[1], 'code': l[0]} for l in settings.LANGUAGES],
                        content_type="application/json")



#!/usr/bin/env python3
import os
import subprocess
from datetime import date
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


def get_version(module):
    version = date.today().strftime('%Y-%m')
    git_tag = "0.0"
    git_commits = "0"
    suffix = "dev"
    try:
        branch = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).rstrip().decode('utf8')
        git_describe = subprocess.check_output(["git", "describe", "--long"]).rstrip().decode('utf8')
        git_tag = git_describe.split('-')[0]
        git_commits = git_describe.split('-')[1]
        if branch == 'master':
            suffix = ''
        else:
            suffix = 'dev'
        print(branch, git_tag, git_commits, suffix)
        version = '{}.{}{}'.format(git_tag, git_commits, suffix)
    except (subprocess.CalledProcessError, OSError) as e:
        print('git not installed', e)
    try:
        fp = open('{}/__init__.py'.format(module), 'w')
        fp.write('__version__ = [{}, {}, "{}"]'.format(git_tag.replace('.', ','), git_commits, suffix))
        fp.close()
    except Exception:
        print('ERROR opening {}/__init__.py'.format(module), os.curdir)
    return version

module = 'geocurrency'

setup(
    name='geocurrency',
    description='Web based services to convert units and currencies.',
    python_requires='>3.8.0',
    version=get_version(module),
    author='Frédéric MEUROU',
    author_email='fm@peabytes.me',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://api.geocurrency.me/swagger/',
    install_requires=[
        "Django~=3.1.0",
        "django-cors-headers~=3.2.0",
        "django-cors-middleware~=1.5.0",
        "django-createsuperuser",
        "django-extensions~=3.0.0",
        "django-filter~=2.3.0",
        "django-redis~=4.12.0",
        "django-sendfile~=0.3.0",
        "djangorestframework~=3.11.0",
        "drf-yasg~=1.17.0",
        "markdown~=3.0",
        "lxml~=4.0",
        "django_redis~=4.0",
        "pysendfile~=2.0",
        "gunicorn",
        # "psycopg2",
        # "mysql",
        "pytz",
        "pycountry",
        "countryinfo~=0.1.0",
        "timezonefinder~=4.4.0",
        "iso4217",
        "forex-python~=1.0",
        "Babel~=2.8",
        "Pint~=0.15",
        "networkx~=2.5.0",
        "sympy~=1.7.0"
    ],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Environment :: Web Environment",
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Framework :: Django :: 3.1",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Office/Business",
        "License :: OSI Approved :: MIT License",
    ],
    py_modules=[
        'geocurrency.core',
        'geocurrency.countries',
        'geocurrency.currencies',
        'geocurrency.rates',
        'geocurrency.units',
        'geocurrency.converters'
    ],
)



# Generated by Django 3.1.2 on 2020-10-17 21:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomUnit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, default=None, max_length=255, null=True)),
                ('unit_system', models.CharField(choices=[('Planck', 'Planck'), ('SI', 'SI'), ('US', 'US'), ('atomic', 'atomic'), ('cgs', 'CGS'), ('imperial', 'imperial'), ('mks', 'mks')], max_length=20)),
                ('code', models.SlugField()),
                ('name', models.CharField(max_length=255, verbose_name='Human readable name')),
                ('relation', models.CharField(max_length=255, verbose_name='Relation to an existing unit')),
                ('symbol', models.CharField(max_length=20, verbose_name='Symbol')),
                ('alias', models.CharField(blank=True, max_length=20, null=True, verbose_name='Alias')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='units', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'key', 'code')},
            },
        ),
    ]



"""
Pelias geocoder
"""
import logging

import requests
from django.conf import settings
from pycountry import countries

from . import Geocoder
from ..settings import *


class PeliasGeocoder(Geocoder):
    """
    Pelias geocoder
    """
    coder_type = 'pelias'
    server_url = None
    key = None

    def __new__(cls, *args, **kwargs):
        """
        Initialize
        """
        return super(Geocoder, cls).__new__(cls)

    def __init__(self, server_url: str = None, key: str = None, *args, **kwargs):
        """
        Init pelias geocoder
        Init: Geocoder('pelias', server_url='serveur URL', key='API key')
        :params server_url: Custom pelias URL, defaults to 'https://api.geocode.earth/v1/search'
        :param key: API key
        """
        self.key = key or settings.GEOCODER_PELIAS_KEY
        try:
            pelias_url = settings.GEOCODER_PELIAS_URL
        except AttributeError:
            pelias_url = GEOCODING_SERVICE_SETTINGS['pelias']['default_url']
        self.server_url = server_url or pelias_url

    def search(self, address: str, language: str = None, bounds: str = None, region: str = None,
               components: str = "") -> dict:
        """
        Search address
        :param address: Address to look for
        :param language: language of the query
        :param bounds: Unused at the moment
        :param region: not used
        :param components: not used
        """
        search_args = {'text': address}
        if self.key:
            search_args['api_key'] = self.key
        try:
            response = requests.get(f'{self.server_url}/search', search_args)
            data = response.json()
            if 'errors' in data:
                logging.error("Invalid request")
                logging.error(data.get('error'))
                return {}
            return data
        except ValueError as e:
            logging.error("Invalid API configuration")
            logging.error(e)
        except IOError as e:
            logging.error("Invalid request")
            logging.error(e)
        return {}

    def reverse(self, lat: str, lng: str) -> dict:
        """
        Reverse search by coordinates
        :param lat: latitude
        :param lng: longitude
        """
        search_args = {
            'point.lat': lat,
            'point.lon': lng,
        }
        if self.key:
            search_args['api_key'] = self.key
        try:
            response = requests.get(f'{self.server_url}/reverse', search_args)
            data = response.json()
            if 'errors' in data:
                logging.error("ERROR - Invalid request")
                logging.error(data.get('error'))
                return {}
            return data
        except ValueError as e:
            logging.error("Invalid API configuration")
            logging.error(e)
        except IOError as e:
            logging.error("Invalid request", e)
        return {}

    def parse_countries(self, data: dict) -> [str]:
        """
        parse response from google service
        :params data: geocoding / reverse geocoding json
        :return: array of alpha2 codes
        """
        alphas = []
        if not data:
            return alphas
        for feature in data.get('features'):
            alphas.append(countries.get(alpha_3=feature.get('properties').get('country_a')).alpha_2)
        return alphas



"""
Forms for Rates module
"""
from django.forms import ModelForm

from .models import Rate


class RateForm(ModelForm):
    """
    Rate form
    """
    class Meta:
        """
        Model form
        """
        model = Rate
        exclude = ['user']
